with Messages,Reps,Reps.Ops,Reps.IO,Rounding,Basics;
use  Messages,Reps,Reps.Ops,Reps.IO;
with Intervals,Intervals.Ops,Intervals.IO;
use  Intervals,Intervals.Ops,Intervals.IO;

procedure Agg is
   subtype Scalar is Interval;
   PDeg: constant Positive := 150;          -- degree of Taylor polynomials

   package Basic is new Basics (PDeg => PDeg, Scalar => Scalar);
   package TC renames Basic.TCoeffs;
   package TC_Ops Renames Basic.TCoeffs_Ops;
   package TC_IO renames Basic.TCoeffs_IO;
   use Basic,TC,TC_Ops,TC_IO;

   M: Boolean;
   B1: constant Rep := -2.5173682746;
   B2: constant Rep := -2.51736827425;
   T1: constant Rep := 6.95;    -- integration time for B1
   T2: constant Rep := 7.8;     -- integration time for B2
   T3: constant Rep := 5.8;     -- integration time for intermediate points
   D:  constant Rep := 3.5e-13; -- delta B
   C1,C2: Rep := B1;
   N: constant Positive := 5;       -- dimension of the space
   B: Scalar;
   C: constant Scalar := Scal(0.37);
   R: constant Scalar := Scal(0.45);
begin
   if FlowOut(C,R) <= Zero then
      Message("not out",Parameter_Error);
   end if;
   if FlowIn(C,R) >= Zero then
      Message("not in",Parameter_Error);
   end if;
   B := Scal(B1);
   M := Integrate(N,PDeg,T1,B,C,R,True,False);
   if M then
      Show("ok down");
   end if;
   B := Scal(B2);
   M := M and Integrate(N,PDeg,T2,B,C,R,False,True);
   if M then
      Show("ok up");
   end if;
   for I in 1 .. 1000 loop
      C1 := C2;
      C2 := C1+D;
      B  := Scal(C1,C2);
      M := M and Integrate(N,PDeg,T3,B,C,R,False,False);
      if M then
         Show("ok",I);
      end if;
   end loop;
   if M then
      Show("ok");
   end if;
end Agg;
