with Messages, Reps.Ops,Rounding;
use Messages, Reps.Ops;

package body Basics is

   Trunc:   constant Boolean := IsNumeric(Scal(Zero));
   Four:    constant Rep     := Two*Two;
   ScalOne: constant Scalar  := Scal(One);
   ScalHalf:constant Scalar  := Scal(Half);
   Scal6th: constant Scalar  := ScalOne/Scal(Six);

   function FirstStep(R: Radius; Beta: Scalar; N: Positive) return Taylor is
      Lambda: Scalar := Scal(8*(N-2)*(N-4));
      U: Taylor := CMonom(0,ScalOne);
      T: Taylor;
      K: Natural;
   begin
      SetComponent(2,ScalHalf*Beta,U);
      for I in 0 .. PDeg/2-2 loop
         K := 2*I;
         T := Lambda*Exp(R,U);
         SetComponent(K+4,Component(K,T)/C(K,N),U);
      end loop;
      return U;
   end FirstStep;

   function SecondStep(R: Radius; D0,D1,D2,D3: Scalar; N: Positive) return Taylor is
      Lambda: Scalar := Scal(8*(N-2)*(N-4));
      U: Taylor := CMonom(0,D0);
      S: Scalar;
   begin
      SetComponent(1,D1,U);
      SetComponent(2,ScalHalf*D2,U);
      SetComponent(3,Scal6th*D3,U);
      for K in 0 .. PDeg-4 loop
         S := Scal(Two*Rep((N-4)*(K+3)*(K+2)*(K+1)))*Component(K+3,U);
         S := S+Scal((N*N-10*N+20)*(K+2)*(K+1))*Component(K+2,U);
         S := S-Scal(Two*Rep((N-2)*(N-4)*(K+1)))*Component(K+1,U);
         if K>0 then
            S := -S+Lambda*Component(K,Exp(R,U));
         else
            S := -S+Lambda*(Component(0,Exp(R,U))-ScalOne);
         end if;
         S := S/Scal((k+4)*(k+3)*(k+2)*(k+1));
         SetComponent(K+4,S,U);
      end loop;
      return U;
   end SecondStep;

   function C(KK,N: Natural) return Scalar is
      K:  constant Scalar := Scal(KK);
      S1: constant Scalar := Scal(One);
      S2: constant Scalar := Scal(Two);
      S3: constant Scalar := Scal(Three);
      S4: constant Scalar := Scal(Two*Two);
   begin
      return (K+S4)*(K+S3)*(K+S2)*(K+S1)+Scal(2*(N-1))*(K+S4)*(K+S3)*(K+S2)+
        Scal((N-1)*(N-3))*(K+S4)*(K+S2);
   end C;

   function CFirst(R: Radius; Beta: Scalar; N: Positive; U: Taylor) return Taylor is
      Lambda: Scalar := Scal(8*(N-2)*(N-4));
      T: Taylor;
      V: Taylor := CMonom(0,ScalOne);
      K: Natural;
   begin
      T := Lambda*Exp(R,U);
      SetComponent(2,ScalHalf*Beta,V);
      for I in 0 .. PDeg/2-2 loop
         K := I*2;
         SetComponent(K+4,Component(K,T)/C(K,N),V);
      end loop;
      SetErrComp(abs(Component(PDeg-2,T))/C(PDeg-2,N)+
                 abs(Component(PDeg,T))/C(PDeg,N)+
                 ErrComp(T)/C(PDeg+2,N),V);
      return V;
   end CFirst;

   function CSecond(R: Radius; D0,D1,D2,D3: Scalar;
                    N: Positive; U: Taylor) return Taylor is
      V: Taylor;
   begin
      V := Scal(8*(N-2)*(N-4))*((Exp(R,U)-CMonom(0,Scal(One))));
      V := Scal(2*(N-2)*(N-4))*U+Dm1(R,V);
      V := Scal(-N*N+10*N-20)*U+Dm1(R,V);
      V := Scal(-2*(N-4))*U+Dm1(R,V);
      V := Dm1(R,V);
      SetComponent(0,D0,V);
      SetComponent(1,D1,V);
      SetComponent(2,ScalHalf*D2,V);
      SetComponent(3,Scal6th*D3,V);
      return V;
   end CSecond;

   function Lip1(R: Radius; N:Positive) return Scalar is
   begin
      return Scal(R**4)/C(0,N);
   end Lip1;

   function Lip2(R,U: Radius; N:Positive) return Scalar is
      RR: Scalar:= Scal(R);
   begin
      return Scal(2*(N-4))*RR+ScalHalf*abs(Scal(N*N-10*N+20))*RR*RR
        +(Scal((N-2)*(N-4))+Scal((N-2)*(N-4))*(Exp(Scal(U))+ScalOne)*RR)
        *RR*RR*RR/Scal(Three);
   end Lip2;

   procedure EndPointU(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                      D0,D1,D2,D3: out Scalar) is
      U0,U1,U2,U3: Taylor;
      R3: Radius := Sup(R);     -- radius for the 3rd derivative
      DR: Radius := (R0-R3)/3.0;
      R2: Radius := R3+DR;      -- radius for the 2nd derivative
      R1: Radius := R2+DR;      -- radius for the 1st derivative
   begin
      U0 := Eps*Ball(R0)+U;
      if R3 >= R0 then Message("Basics.EndPointU",Parameter_Error); end if;
      D0 := Val(R0,R,U)+Scal(4)*Log(R);
      U1 := D(R0,R1,U);         -- derivatives of U
      U2 := D(R1,R2,U1);
      U3 := D(R2,R3,U2);
      D1 := R*Val(R1,R,U1)+Scal(4);
      D2 := R*Val(R1,R,U1)+R*R*Val(R2,R,U2);
      D3 := R*Val(R1,R,U1)+Scal(3)*R*R*Val(R2,R,U2)+R*R*R*Val(R3,R,U3);
   end EndPointU;

   procedure EndPointW(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                      D0,D1,D2,D3: out Scalar) is
      U0,U1,U2,U3: Taylor;
      R3: Radius := Sup(R);     -- radius for the 3rd derivative
      DR: Radius := (R0-R3)/3.0;
      R2: Radius := R3+DR;      -- radius for the 2nd derivative
      R1: Radius := R2+DR;      -- radius for the 1st derivative
   begin
      U0 := Eps*Ball(R0)+U;
      if R3 >= R0 then Message("Basics.EndPointW",Parameter_Error); end if;
      D0 := Val(R0,R,U);
      U1 := D(R0,R1,U);         -- derivatives of U
      U2 := D(R1,R2,U1);
      U3 := D(R2,R3,U2);
      D1 := Val(R1,R,U1);
      D2 := Val(R2,R,U2);
      D3 := Val(R3,R,U3);
   end EndPointW;

   procedure SetUp is
      EV: SMatrix(1 .. 4,1 .. 4);
      R33: constant Scalar := Sqrt(Scal(33));
      R2p: constant Scalar := Sqrt(Scal(4)*R33+Scal(13));
      R2m: constant Scalar := Sqrt(Scal(4)*R33-Scal(13));
      S0:  constant Scalar := Scal(0);
      S1:  constant Scalar := Scal(1);
      S2:  constant Scalar := Scal(2);
      S3:  constant Scalar := Scal(3);
      S4:  constant Scalar := Scal(4);
      S6:  constant Scalar := Scal(6);
      S7:  constant Scalar := Scal(7);
      S12: constant Scalar := Scal(12);
      S21: constant Scalar := Scal(21);
      S48: constant Scalar := Scal(48);
      S69: constant Scalar := Scal(69);
   begin
      EvUn := (R2p-S1)/s2;  -- positive eigenvalue
      EV(1,1) := (S3-R33)*(R2p-S1)/(S12*Sqr(S1+R2p)); -- eigenvectors
      EV(2,1) := S4/Sqr(S1+R2p);
      EV(3,1) := -S2/(S1+R2p);
      EV(4,1) := S1;
      EV(1,2) := (R33-S3)*(R2p+S1)/(S12*Sqr(R2p-S1));
      EV(2,2) := S4/Sqr(S1-R2p);
      EV(3,2) := S2/(R2p-S1);
      EV(4,2) := S1;
      EV(1,3) := (S69-R33)/(S48*Sqr(R33-S3));
      EV(2,3) := (S7-S2*R33)/(S2*Sqr(R33-S3));
      EV(3,3) := S1/(S6-S2*R33);
      EV(4,3) := S1;
      EV(1,4) := (R33-S21)*R2m/(S48*Sqr(R33-S3));
      EV(2,4) := -R2m/(S2*Sqr(R33-S3));
      EV(3,4) := R2m/(S2*R33-S6);
      EV(4,4) := S0;
      for I in 1 .. 2 loop  -- normalizes ev1 and ev2
         C1 := S0;
         for J in 1 .. 4 loop
            C1 := C1+Sqr(EV(J,I));
         end loop;
         C1 := Sqrt(C1);
         for J in 1 .. 4 loop
            EV(J,I) := EV(J,I)/C1;
         end loop;
      end loop;
      C1 := S0;
      for I in 3 .. 4 loop  -- normalizes ev3 and ev4
         for J in 1 .. 4 loop
            C1 := C1+Sqr(EV(J,I));
         end loop;
      end loop;
      C1 := Sqrt(C1);
      for I in 3 .. 4 loop
         for J in 1 .. 4 loop
            EV(J,I) := EV(J,I)/C1;
         end loop;
      end loop;
      C1 := S0;
      for I in 1 .. 4 loop -- compute C1
        C1 := C1 + Sqr(EV(1,I));
      end loop;
      C1 := Sqrt(C1);
      EVI := EV;
      Invert(EVI);
   end SetUp;

   function ScalarProduct(U,V: SVector) return Scalar is
      A,B: SVector(1 .. 4);
   begin
      A := EVI*U;
      B := EVI*V;
      return A(1)*B(1)+A(2)*B(2)+A(3)*B(3)+A(4)*B(4);
   end ScalarProduct;

   function Norm(U: SVector) return Scalar is
   begin
      return Sqrt(abs(ScalarProduct(U,U)));
   end Norm;

   function Norm(U1,U2,U3,U4: Scalar) return Scalar is
      U: SVector(1 .. 4);
   begin
      U(1) := U1; U(2) := U2; U(3) := U3; U(4) := U4;
      return Norm(U);
   end Norm;

   function Alpha(U1,U2,U3,U4: Scalar) return Scalar is
      U,V: SVector(1 .. 4);
   begin
      U(1) := U1; U(2) := U2; U(3) := U3; U(4) := U4;
      V := EVI*U;
      return V(2);
   end Alpha;

   function FlowIn(Sp,Nm: Scalar) return Rep is
   begin
      return  Sup(Scal(-Half)*(Scal(One)-Sqr(Sp))+Scal(12)*Exp(C1*Nm)*C1*C1*Nm);
   end FlowIn;

   function FlowOut(Sp,Nm: Scalar) return Rep is
   begin
      return Inf(EvUn*Sqr(Sp)-Scal(12)*Exp(C1*Nm)*C1*C1*Nm);
   end FlowOut;

   function Integrate(N,PDeg: Positive; DT:Rep;
                      Beta,C,R: Scalar; Down,Up: Boolean) return Boolean is

      U,V: Taylor;
      R0: Radius := 1.2; -- radius for the first step
      R1: Radius := 0.3; -- radius for the following steps
      MaxRepet: constant Natural := 40;
      Nm,Sp,Eps,D0,D1,D2,D3: Scalar;
      T: Rep := Zero;
   begin
      U := FirstStep(R0,Beta,N);
      V := U-CFirst(R0,Beta,N,U);
      SetComponent(0,Scal(Zero),V);  -- resets components that are 0 by definition
      SetComponent(2,Scal(Zero),V);
      Eps := Widen(Norm(R0,V)/(Scal(One)-Lip1(R0,N)));
      EndPointU(R0,Scal(One),Eps,U,D0,D1,D2,D3);
      while T < DT loop
         U := SecondStep(R1,D0,D1,D2,D3,N);
         V := U-CSecond(R1,D0,D1,D2,D3,N,U);
         SetComponent(0,Scal(Zero),V);  -- resets components that are 0 by definition
         SetComponent(1,Scal(Zero),V);
         SetComponent(2,Scal(Zero),V);
         SetComponent(3,Scal(Zero),V);
         Eps := Norm(R1,V);
         Eps := Widen(Eps/(Scal(One)-Lip2(R1,Sup(Eps+Norm(R1,U)),N)));
         EndPointW(R1,Scal(Fourth),Eps,U,D0,D1,D2,D3);
         T := T+Fourth;
      end loop;
      EndPointW(R1,Scal(DT-T+Fourth),Eps,U,D0,D1,D2,D3);
      Nm := Norm(D0,D1,D2,D3);
      Sp := Alpha(D0,D1,D2,D3)/Nm;
      if Down and Sup(Sp) > Inf(-C) then
         Show(Sp);
         Show(C);
         Show("not down");
         return False;
      end if;
      if Up and Inf(Sp) < Sup(C) then
         Show("not up");
         return False;
      end if;
      if Sup(Nm) > Inf(R) then
         Show(Nm);
         Show(R);
         Show("not in ball");
         return False;
      end if;
      return True;
   end Integrate;

begin
   SetUp;
end Basics;
