with Ada.Text_IO,Reps,Taylors,Taylors.Ops,Taylors.IO;
with SLinear,SLinear.IO,SLinear.ops;
use  Ada.Text_IO, Reps;

generic

  PDeg: in Positive;
  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Approx(S: Scalar) return Rep is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Scal(R1,R2: Rep) return Scalar is <>;
  with function Ball(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function Inf(S: Scalar) return Rep is <>;
  with function Sup(S: Scalar) return Rep is <>;
  with function InfAbs(S: Scalar) return Rep is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with function "*"(R: Rep; S: Scalar) return Scalar is <>;
  with function "/"(S: Scalar; R: Rep) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Widen(S: Scalar) return Scalar is <>;
  with procedure Show(S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Get(F: in File_Type; S: out Scalar; Decimal: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := True) is <>;

package Basics is

  package TCoeffs is new Taylors(PDeg => PDeg, Scalar => Scalar);
  package TCoeffs_Ops is new TCoeffs.Ops;
  package TCoeffs_IO is new TCoeffs.IO ;
  use TCoeffs, TCoeffs_Ops, TCoeffs_IO;
  package QSLinear is new SLinear(Scalar => Scalar);
  package QSLinear_Ops is new QSLinear.Ops;
  package QSLinear_IO is new QSLinear.IO;
  use QSLinear, QSLinear_Ops, QSLinear_IO;

  function FirstStep(R: Radius; Beta: Scalar; N: Positive) return Taylor;
  function SecondStep(R: Radius; D0,D1,D2,D3: Scalar; N: Positive) return Taylor;
  function CFirst(R: Radius; Beta: Scalar; N: Positive; U: Taylor) return Taylor;
  function CSecond(R: Radius; D0,D1,D2,D3: Scalar;
                   N: Positive; U: Taylor) return Taylor;
  function Lip1(R: Radius; N:Positive) return Scalar;  -- Lipschitz constant of CFirst
  function Lip2(R,U: Radius; N:Positive) return Scalar;-- Lipschitz constant of CSecond
  procedure EndPointU(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                     D0,D1,D2,D3: out Scalar);
  procedure EndPointW(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                                                       D0,D1,D2,D3: out Scalar);
  procedure SetUp;
  function ScalarProduct(U,V: SVector) return Scalar;
  function Norm(U: SVector) return Scalar;
  function Norm(U1,U2,U3,U4: Scalar) return Scalar;
  function Alpha(U1,U2,U3,U4: Scalar) return Scalar;
  function FlowIn(Sp,Nm: Scalar) return Rep;
  function FlowOut(Sp,Nm: Scalar) return Rep;
  function Integrate(N,PDeg: Positive; DT:Rep;
                      Beta,C,R: Scalar; Down,Up: Boolean) return Boolean;

private
  function C(KK,N: Natural) return Scalar;
  EVI: SMatrix(1 .. 4,1 .. 4);
  EvUn,C1,C2: Scalar;
end Basics;
