with Reps;
use Reps;

generic

  PDeg: in Natural;
  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function Approx(S: Scalar) return Rep is <>;

package Taylors is

  pragma Pure;

  type Taylor is private;

  function IsScalar(P: Taylor) return Boolean;                           -- False
  function CMonom(K: Natural; S: Scalar) return Taylor;                  -- S*t^k
  function EMonom(S: Scalar) return Taylor;                              -- S*unit_error
  function Component(K: Natural; P: Taylor) return Scalar;               -- P[K]
  procedure SetComponent(K: in Natural; S: in Scalar; P: in out Taylor); -- P[D] := S
  function ErrComp(P: Taylor) return Scalar;                             -- error
  procedure SetErrComp(S: in Scalar; P: in out Taylor);                  -- error := |S|
  function IsZero(P: Taylor) return Boolean;                             -- True if P=0
  function ZeroErr(P: Taylor) return Boolean;                            -- True if error=0

  pragma Inline (Component,SetComponent,ErrComp,SetErrComp);

private

  Size: constant Integer := PDeg+1;

  subtype Power is Integer range 0 .. PDeg;
  subtype Index is Integer range 0 .. Size;
  type ScalVec is array(Integer range <>) of Scalar;
  type Taylor is new ScalVec(Index);             -- last "coeff" is higher order error

  ScalZero:  constant Scalar := Scal(Zero);
  ScalOne:  constant Scalar := Scal(One);
  Trunc: constant Boolean := IsNumeric(ScalZero);

end Taylors;
